@@ -3,12 +3,12 @@
  * index.c
  *	  code to create and destroy POSTGRES index relations
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/index.c,v 1.219 2003/09/29 00:05:24 petere Exp $
+ *	  $PostgreSQL: pgsql/src/backend/catalog/index.c,v 1.244.4.1 2005/06/25 16:54:00 tgl Exp $
  *
  *
  * INTERFACE ROUTINES
@@ -25,7 +25,6 @@
 
 #include "access/genam.h"
 #include "access/heapam.h"
-#include "access/istrat.h"
 #include "bootstrap/bootstrap.h"
 #include "catalog/catalog.h"
 #include "catalog/catname.h"
@@ -34,20 +33,15 @@
 #include "catalog/index.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_constraint.h"
-#include "catalog/pg_index.h"
 #include "catalog/pg_opclass.h"
-#include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
 #include "executor/executor.h"
 #include "miscadmin.h"
 #include "optimizer/clauses.h"
-#include "optimizer/prep.h"
 #include "parser/parse_expr.h"
-#include "parser/parse_func.h"
 #include "storage/sinval.h"
 #include "storage/smgr.h"
 #include "utils/builtins.h"
-#include "utils/catcache.h"
 #include "utils/fmgroids.h"
 #include "utils/inval.h"
 #include "utils/lsyscache.h"
@@ -55,14 +49,6 @@
 #include "utils/syscache.h"
 
 
-/*
- * macros used in guessing how many tuples are on a page.
- */
-#define AVG_ATTR_SIZE 8
-#define NTUPLES_PER_PAGE(natts) \
-	((BLCKSZ - MAXALIGN(sizeof(PageHeaderData))) / \
-	((natts) * AVG_ATTR_SIZE + MAXALIGN(sizeof(HeapTupleHeaderData))))
-
 /* non-export function prototypes */
 static TupleDesc ConstructTupleDescriptor(Relation heapRelation,
 						 IndexInfo *indexInfo,
@@ -89,7 +75,7 @@ ConstructTupleDescriptor(Relation heapRelation,
 						 Oid *classObjectId)
 {
 	int			numatts = indexInfo->ii_NumIndexAttrs;
-	List	   *indexprs = indexInfo->ii_Expressions;
+	ListCell   *indexpr_item = list_head(indexInfo->ii_Expressions);
 	TupleDesc	heapTupDesc;
 	TupleDesc	indexTupDesc;
 	int			natts;			/* #atts in heap rel --- for error checks */
@@ -154,7 +140,7 @@ ConstructTupleDescriptor(Relation heapRelation,
 			 */
 			to->attnum = i + 1;
 
-			to->attstattarget = 0;
+			to->attstattarget = -1;
 			to->attcacheoff = -1;
 			to->attnotnull = false;
 			to->atthasdef = false;
@@ -166,10 +152,10 @@ ConstructTupleDescriptor(Relation heapRelation,
 			/* Expressional index */
 			Node	   *indexkey;
 
-			if (indexprs == NIL)	/* shouldn't happen */
+			if (indexpr_item == NULL)	/* shouldn't happen */
 				elog(ERROR, "too few entries in indexprs list");
-			indexkey = (Node *) lfirst(indexprs);
-			indexprs = lnext(indexprs);
+			indexkey = (Node *) lfirst(indexpr_item);
+			indexpr_item = lnext(indexpr_item);
 
 			/*
 			 * Make the attribute's name "pg_expresssion_nnn" (maybe think
@@ -198,6 +184,7 @@ ConstructTupleDescriptor(Relation heapRelation,
 			to->attbyval = typeTup->typbyval;
 			to->attstorage = typeTup->typstorage;
 			to->attalign = typeTup->typalign;
+			to->attstattarget = -1;
 			to->attcacheoff = -1;
 			to->atttypmod = -1;
 			to->attislocal = true;
@@ -398,13 +385,14 @@ UpdateIndexRelation(Oid indexoid,
 		exprsDatum = (Datum) 0;
 
 	/*
-	 * Convert the index predicate (if any) to a text datum
+	 * Convert the index predicate (if any) to a text datum.  Note we
+	 * convert implicit-AND format to normal explicit-AND for storage.
 	 */
 	if (indexInfo->ii_Predicate != NIL)
 	{
 		char	   *predString;
 
-		predString = nodeToString(indexInfo->ii_Predicate);
+		predString = nodeToString(make_ands_explicit(indexInfo->ii_Predicate));
 		predDatum = DirectFunctionCall1(textin,
 										CStringGetDatum(predString));
 		pfree(predString);
@@ -466,10 +454,12 @@ index_create(Oid heapRelationId,
 			 const char *indexRelationName,
 			 IndexInfo *indexInfo,
 			 Oid accessMethodObjectId,
+			 Oid tableSpaceId,
 			 Oid *classObjectId,
 			 bool primary,
 			 bool isconstraint,
-			 bool allow_system_table_mods)
+			 bool allow_system_table_mods,
+			 bool skip_build)
 {
 	Relation	heapRelation;
 	Relation	indexRelation;
@@ -508,8 +498,10 @@ index_create(Oid heapRelationId,
 	 * We cannot allow indexing a shared relation after initdb (because
 	 * there's no way to make the entry in other databases' pg_class).
 	 * Unfortunately we can't distinguish initdb from a manually started
-	 * standalone backend.	However, we can at least prevent this mistake
-	 * under normal multi-user operation.
+	 * standalone backend (toasting of shared rels happens after the
+	 * bootstrap phase, so checking IsBootstrapProcessingMode() won't
+	 * work).  However, we can at least prevent this mistake under normal
+	 * multi-user operation.
 	 */
 	if (shared_relation && IsUnderPostmaster)
 		ereport(ERROR,
@@ -536,9 +528,10 @@ index_create(Oid heapRelationId,
 	 */
 	indexRelation = heap_create(indexRelationName,
 								namespaceId,
+								tableSpaceId,
 								indexTupDesc,
+								RELKIND_INDEX,
 								shared_relation,
-								true,
 								allow_system_table_mods);
 
 	/* Fetch the relation OID assigned by heap_create */
@@ -707,33 +700,51 @@ index_create(Oid heapRelationId,
 	}
 
 	/*
-	 * Fill in the index strategy structure with information from the
-	 * catalogs.  First we must advance the command counter so that we
-	 * will see the newly-entered index catalog tuples.
+	 * Advance the command counter so that we can see the newly-entered
+	 * catalog tuples for the index.
 	 */
 	CommandCounterIncrement();
 
-	RelationInitIndexAccessInfo(indexRelation);
+	/*
+	 * In bootstrap mode, we have to fill in the index strategy structure
+	 * with information from the catalogs.  If we aren't bootstrapping,
+	 * then the relcache entry has already been rebuilt thanks to sinval
+	 * update during CommandCounterIncrement.
+	 */
+	if (IsBootstrapProcessingMode())
+		RelationInitIndexAccessInfo(indexRelation);
+	else
+		Assert(indexRelation->rd_indexcxt != NULL);
 
 	/*
 	 * If this is bootstrap (initdb) time, then we don't actually fill in
 	 * the index yet.  We'll be creating more indexes and classes later,
 	 * so we delay filling them in until just before we're done with
-	 * bootstrapping.  Otherwise, we call the routine that constructs the
-	 * index.
+	 * bootstrapping.  Similarly, if the caller specified skip_build then
+	 * filling the index is delayed till later (ALTER TABLE can save work
+	 * in some cases with this).  Otherwise, we call the AM routine that
+	 * constructs the index.
 	 *
-	 * In normal processing mode, the heap and index relations are closed by
-	 * index_build() --- but we continue to hold the ShareLock on the heap
-	 * and the exclusive lock on the index that we acquired above, until
-	 * end of transaction.
+	 * In normal processing mode, the heap and index relations are closed,
+	 * but we continue to hold the ShareLock on the heap and the exclusive
+	 * lock on the index that we acquired above, until end of transaction.
 	 */
 	if (IsBootstrapProcessingMode())
 	{
 		index_register(heapRelationId, indexoid, indexInfo);
 		/* XXX shouldn't we close the heap and index rels here? */
 	}
+	else if (skip_build)
+	{
+		/* caller is responsible for filling the index later on */
+		relation_close(indexRelation, NoLock);
+		heap_close(heapRelation, NoLock);
+	}
 	else
+	{
 		index_build(heapRelation, indexRelation, indexInfo);
+		/* index_build closes the passed rels */
+	}
 
 	return indexoid;
 }
@@ -752,9 +763,7 @@ index_drop(Oid indexId)
 	Relation	userIndexRelation;
 	Relation	indexRelation;
 	HeapTuple	tuple;
-	int			i;
-
-	Assert(OidIsValid(indexId));
+	bool		hasexprs;
 
 	/*
 	 * To drop an index safely, we must grab exclusive lock on its parent
@@ -763,7 +772,7 @@ index_drop(Oid indexId)
 	 * backend might be in the midst of devising a query plan that will
 	 * use the index.  The parser and planner take care to hold an
 	 * appropriate lock on the parent table while working, but having them
-	 * hold locks on all the indexes too seems overly complex.	We do grab
+	 * hold locks on all the indexes too seems overly expensive.  We do grab
 	 * exclusive lock on the index too, just to be safe. Both locks must
 	 * be held till end of transaction, else other backends will still see
 	 * this index in pg_index.
@@ -775,17 +784,25 @@ index_drop(Oid indexId)
 	LockRelation(userIndexRelation, AccessExclusiveLock);
 
 	/*
-	 * fix RELATION relation
+	 * flush buffer cache and schedule physical removal of the file
 	 */
-	DeleteRelationTuple(indexId);
+	FlushRelationBuffers(userIndexRelation, (BlockNumber) 0);
+
+	RelationOpenSmgr(userIndexRelation);
+	smgrscheduleunlink(userIndexRelation->rd_smgr,
+					   userIndexRelation->rd_istemp);
 
 	/*
-	 * fix ATTRIBUTE relation
+	 * Close and flush the index's relcache entry, to ensure relcache
+	 * doesn't try to rebuild it while we're deleting catalog entries. We
+	 * keep the lock though.
 	 */
-	DeleteAttributeTuples(indexId);
+	index_close(userIndexRelation);
+
+	RelationForgetRelation(indexId);
 
 	/*
-	 * fix INDEX relation
+	 * fix INDEX relation, and check for expressional index
 	 */
 	indexRelation = heap_openr(IndexRelationName, RowExclusiveLock);
 
@@ -795,19 +812,29 @@ index_drop(Oid indexId)
 	if (!HeapTupleIsValid(tuple))
 		elog(ERROR, "cache lookup failed for index %u", indexId);
 
+	hasexprs = !heap_attisnull(tuple, Anum_pg_index_indexprs);
+
 	simple_heap_delete(indexRelation, &tuple->t_self);
 
 	ReleaseSysCache(tuple);
 	heap_close(indexRelation, RowExclusiveLock);
 
 	/*
-	 * flush buffer cache and physically remove the file
+	 * if it has any expression columns, we might have stored statistics
+	 * about them.
 	 */
-	i = FlushRelationBuffers(userIndexRelation, (BlockNumber) 0);
-	if (i < 0)
-		elog(ERROR, "FlushRelationBuffers returned %d", i);
+	if (hasexprs)
+		RemoveStatistics(indexId, 0);
 
-	smgrunlink(DEFAULT_SMGR, userIndexRelation);
+	/*
+	 * fix ATTRIBUTE relation
+	 */
+	DeleteAttributeTuples(indexId);
+
+	/*
+	 * fix RELATION relation
+	 */
+	DeleteRelationTuple(indexId);
 
 	/*
 	 * We are presently too lazy to attempt to compute the new correct
@@ -817,15 +844,12 @@ index_drop(Oid indexId)
 	 * owning relation to ensure other backends update their relcache
 	 * lists of indexes.
 	 */
-	CacheInvalidateRelcache(heapId);
+	CacheInvalidateRelcache(userHeapRelation);
 
 	/*
-	 * Close rels, but keep locks
+	 * Close owning rel, but keep lock
 	 */
-	index_close(userIndexRelation);
 	heap_close(userHeapRelation, NoLock);
-
-	RelationForgetRelation(indexId);
 }
 
 /* ----------------------------------------------------------------
@@ -901,7 +925,7 @@ FormIndexDatum(IndexInfo *indexInfo,
 			   Datum *datum,
 			   char *nullv)
 {
-	List	   *indexprs;
+	ListCell   *indexpr_item;
 	int			i;
 
 	if (indexInfo->ii_Expressions != NIL &&
@@ -914,7 +938,7 @@ FormIndexDatum(IndexInfo *indexInfo,
 		/* Check caller has set up context correctly */
 		Assert(GetPerTupleExprContext(estate)->ecxt_scantuple->val == heapTuple);
 	}
-	indexprs = indexInfo->ii_ExpressionsState;
+	indexpr_item = list_head(indexInfo->ii_ExpressionsState);
 
 	for (i = 0; i < indexInfo->ii_NumIndexAttrs; i++)
 	{
@@ -935,19 +959,19 @@ FormIndexDatum(IndexInfo *indexInfo,
 			/*
 			 * Index expression --- need to evaluate it.
 			 */
-			if (indexprs == NIL)
+			if (indexpr_item == NULL)
 				elog(ERROR, "wrong number of index expressions");
-			iDatum = ExecEvalExprSwitchContext((ExprState *) lfirst(indexprs),
+			iDatum = ExecEvalExprSwitchContext((ExprState *) lfirst(indexpr_item),
 										  GetPerTupleExprContext(estate),
 											   &isNull,
 											   NULL);
-			indexprs = lnext(indexprs);
+			indexpr_item = lnext(indexpr_item);
 		}
 		datum[i] = iDatum;
 		nullv[i] = (isNull) ? 'n' : ' ';
 	}
 
-	if (indexprs != NIL)
+	if (indexpr_item != NULL)
 		elog(ERROR, "wrong number of index expressions");
 }
 
@@ -978,7 +1002,7 @@ setRelhasindex(Oid relid, bool hasindex, bool isprimary, Oid reltoastidxid)
 
 	/*
 	 * Find the tuple to update in pg_class.  In bootstrap mode we can't
-	 * use heap_update, so cheat and overwrite the tuple in-place.  In
+	 * use heap_update, so cheat and overwrite the tuple in-place.	In
 	 * normal processing, make a copy to scribble on.
 	 */
 	pg_class = heap_openr(RelationRelationName, RowExclusiveLock);
@@ -993,10 +1017,10 @@ setRelhasindex(Oid relid, bool hasindex, bool isprimary, Oid reltoastidxid)
 	{
 		ScanKeyData key[1];
 
-		ScanKeyEntryInitialize(&key[0], 0,
-							   ObjectIdAttributeNumber,
-							   F_OIDEQ,
-							   ObjectIdGetDatum(relid));
+		ScanKeyInit(&key[0],
+					ObjectIdAttributeNumber,
+					BTEqualStrategyNumber, F_OIDEQ,
+					ObjectIdGetDatum(relid));
 
 		pg_class_scan = heap_beginscan(pg_class, SnapshotNow, 1, key);
 		tuple = heap_getnext(pg_class_scan, ForwardScanDirection);
@@ -1044,7 +1068,7 @@ setRelhasindex(Oid relid, bool hasindex, bool isprimary, Oid reltoastidxid)
 		/* Send out shared cache inval if necessary */
 		if (!IsBootstrapProcessingMode())
 			CacheInvalidateHeapTuple(pg_class, tuple);
-		BufferSync();
+		BufferSync(-1, -1);
 	}
 	else if (dirty)
 	{
@@ -1056,7 +1080,7 @@ setRelhasindex(Oid relid, bool hasindex, bool isprimary, Oid reltoastidxid)
 	else
 	{
 		/* no need to change tuple, but force relcache rebuild anyway */
-		CacheInvalidateRelcache(relid);
+		CacheInvalidateRelcacheByTuple(tuple);
 	}
 
 	if (!pg_class_scan)
@@ -1076,10 +1100,11 @@ void
 setNewRelfilenode(Relation relation)
 {
 	Oid			newrelfilenode;
+	RelFileNode newrnode;
+	SMgrRelation srel;
 	Relation	pg_class;
 	HeapTuple	tuple;
 	Form_pg_class rd_rel;
-	RelationData workrel;
 
 	/* Can't change relfilenode for nailed tables (indexes ok though) */
 	Assert(!relation->rd_isnailed ||
@@ -1091,13 +1116,13 @@ setNewRelfilenode(Relation relation)
 	newrelfilenode = newoid();
 
 	/*
-	 * Find the pg_class tuple for the given relation.  This is not used
+	 * Find the pg_class tuple for the given relation.	This is not used
 	 * during bootstrap, so okay to use heap_update always.
 	 */
 	pg_class = heap_openr(RelationRelationName, RowExclusiveLock);
 
 	tuple = SearchSysCacheCopy(RELOID,
-							   ObjectIdGetDatum(RelationGetRelid(relation)),
+							ObjectIdGetDatum(RelationGetRelid(relation)),
 							   0, 0, 0);
 	if (!HeapTupleIsValid(tuple))
 		elog(ERROR, "could not find tuple for relation %u",
@@ -1106,17 +1131,21 @@ setNewRelfilenode(Relation relation)
 
 	/* create another storage file. Is it a little ugly ? */
 	/* NOTE: any conflict in relfilenode value will be caught here */
-	memcpy((char *) &workrel, relation, sizeof(RelationData));
-	workrel.rd_fd = -1;
-	workrel.rd_node.relNode = newrelfilenode;
-	heap_storage_create(&workrel);
-	smgrclose(DEFAULT_SMGR, &workrel);
+	newrnode = relation->rd_node;
+	newrnode.relNode = newrelfilenode;
+
+	srel = smgropen(newrnode);
+	smgrcreate(srel, relation->rd_istemp, false);
+	smgrclose(srel);
 
 	/* schedule unlinking old relfilenode */
-	smgrunlink(DEFAULT_SMGR, relation);
+	RelationOpenSmgr(relation);
+	smgrscheduleunlink(relation->rd_smgr, relation->rd_istemp);
 
 	/* update the pg_class row */
 	rd_rel->relfilenode = newrelfilenode;
+	rd_rel->relpages = 0;		/* it's empty until further notice */
+	rd_rel->reltuples = 0;
 	simple_heap_update(pg_class, &tuple->t_self, tuple);
 	CatalogUpdateIndexes(pg_class, tuple);
 
@@ -1134,7 +1163,7 @@ setNewRelfilenode(Relation relation)
  *
  * Update pg_class' relpages and reltuples statistics for the given relation
  * (which can be either a table or an index).  Note that this is not used
- * in the context of VACUUM.
+ * in the context of VACUUM, only CREATE INDEX.
  * ----------------
  */
 void
@@ -1171,15 +1200,16 @@ UpdateStats(Oid relid, double reltuples)
 
 	/*
 	 * Find the tuple to update in pg_class.  Normally we make a copy of
-	 * the tuple using the syscache, modify it, and apply heap_update.
-	 * But in bootstrap mode we can't use heap_update, so we cheat and
-	 * overwrite the tuple in-place.
+	 * the tuple using the syscache, modify it, and apply heap_update. But
+	 * in bootstrap mode we can't use heap_update, so we cheat and
+	 * overwrite the tuple in-place.  (Note: as of PG 8.0 this isn't called
+	 * during bootstrap, but leave the code here for possible future use.)
 	 *
-	 * We also must cheat if reindexing pg_class itself, because the
-	 * target index may presently not be part of the set of indexes that
+	 * We also must cheat if reindexing pg_class itself, because the target
+	 * index may presently not be part of the set of indexes that
 	 * CatalogUpdateIndexes would update (see reindex_relation).  In this
 	 * case the stats updates will not be WAL-logged and so could be lost
-	 * in a crash.  This seems OK considering VACUUM does the same thing.
+	 * in a crash.	This seems OK considering VACUUM does the same thing.
 	 */
 	pg_class = heap_openr(RelationRelationName, RowExclusiveLock);
 
@@ -1196,10 +1226,10 @@ UpdateStats(Oid relid, double reltuples)
 	{
 		ScanKeyData key[1];
 
-		ScanKeyEntryInitialize(&key[0], 0,
-							   ObjectIdAttributeNumber,
-							   F_OIDEQ,
-							   ObjectIdGetDatum(relid));
+		ScanKeyInit(&key[0],
+					ObjectIdAttributeNumber,
+					BTEqualStrategyNumber, F_OIDEQ,
+					ObjectIdGetDatum(relid));
 
 		pg_class_scan = heap_beginscan(pg_class, SnapshotNow, 1, key);
 		tuple = heap_getnext(pg_class_scan, ForwardScanDirection);
@@ -1209,45 +1239,14 @@ UpdateStats(Oid relid, double reltuples)
 		elog(ERROR, "could not find tuple for relation %u", relid);
 	rd_rel = (Form_pg_class) GETSTRUCT(tuple);
 
-	/*
-	 * Figure values to insert.
-	 *
-	 * If we found zero tuples in the scan, do NOT believe it; instead put a
-	 * bogus estimate into the statistics fields.  Otherwise, the common
-	 * pattern "CREATE TABLE; CREATE INDEX; insert data" leaves the table
-	 * with zero size statistics until a VACUUM is done.  The optimizer
-	 * will generate very bad plans if the stats claim the table is empty
-	 * when it is actually sizable.  See also CREATE TABLE in heap.c.
-	 *
-	 * Note: this path is also taken during bootstrap, because bootstrap.c
-	 * passes reltuples = 0 after loading a table.	We have to estimate
-	 * some number for reltuples based on the actual number of pages.
-	 */
-	relpages = RelationGetNumberOfBlocks(whichRel);
-
-	if (reltuples == 0)
-	{
-		if (relpages == 0)
-		{
-			/* Bogus defaults for a virgin table, same as heap.c */
-			reltuples = 1000;
-			relpages = 10;
-		}
-		else if (whichRel->rd_rel->relkind == RELKIND_INDEX && relpages <= 2)
-		{
-			/* Empty index, leave bogus defaults in place */
-			reltuples = 1000;
-		}
-		else
-			reltuples = ((double) relpages) * NTUPLES_PER_PAGE(whichRel->rd_rel->relnatts);
-	}
-
 	/*
 	 * Update statistics in pg_class, if they changed.	(Avoiding an
 	 * unnecessary update is not just a tiny performance improvement; it
 	 * also reduces the window wherein concurrent CREATE INDEX commands
 	 * may conflict.)
 	 */
+	relpages = RelationGetNumberOfBlocks(whichRel);
+
 	if (rd_rel->relpages != (int32) relpages ||
 		rd_rel->reltuples != (float4) reltuples)
 	{
@@ -1419,7 +1418,7 @@ IndexBuildHeapScan(Relation heapRelation,
 	scan = heap_beginscan(heapRelation, /* relation */
 						  snapshot,		/* seeself */
 						  0,	/* number of keys */
-						  (ScanKey) NULL);		/* scan key */
+						  NULL);	/* scan key */
 
 	reltuples = 0;
 
@@ -1436,18 +1435,16 @@ IndexBuildHeapScan(Relation heapRelation,
 		{
 			/* do our own time qual check */
 			bool		indexIt;
-			uint16		sv_infomask;
 
 			/*
-			 * HeapTupleSatisfiesVacuum may update tuple's hint status
-			 * bits. We could possibly get away with not locking the
-			 * buffer here, since caller should hold ShareLock on the
-			 * relation, but let's be conservative about it.
+			 * We could possibly get away with not locking the buffer here,
+			 * since caller should hold ShareLock on the relation, but let's
+			 * be conservative about it.
 			 */
 			LockBuffer(scan->rs_cbuf, BUFFER_LOCK_SHARE);
-			sv_infomask = heapTuple->t_data->t_infomask;
 
-			switch (HeapTupleSatisfiesVacuum(heapTuple->t_data, OldestXmin))
+			switch (HeapTupleSatisfiesVacuum(heapTuple->t_data, OldestXmin,
+											 scan->rs_cbuf))
 			{
 				case HEAPTUPLE_DEAD:
 					indexIt = false;
@@ -1478,7 +1475,7 @@ IndexBuildHeapScan(Relation heapRelation,
 					 * system catalogs before committing.
 					 */
 					if (!TransactionIdIsCurrentTransactionId(
-							  HeapTupleHeaderGetXmin(heapTuple->t_data))
+							   HeapTupleHeaderGetXmin(heapTuple->t_data))
 						&& !IsSystemRelation(heapRelation))
 						elog(ERROR, "concurrent insert in progress");
 					indexIt = true;
@@ -1496,7 +1493,7 @@ IndexBuildHeapScan(Relation heapRelation,
 					 * system catalogs before committing.
 					 */
 					if (!TransactionIdIsCurrentTransactionId(
-							  HeapTupleHeaderGetXmax(heapTuple->t_data))
+							   HeapTupleHeaderGetXmax(heapTuple->t_data))
 						&& !IsSystemRelation(heapRelation))
 						elog(ERROR, "concurrent delete in progress");
 					indexIt = true;
@@ -1508,10 +1505,6 @@ IndexBuildHeapScan(Relation heapRelation,
 					break;
 			}
 
-			/* check for hint-bit update by HeapTupleSatisfiesVacuum */
-			if (sv_infomask != heapTuple->t_data->t_infomask)
-				SetBufferCommitInfoNeedsSave(scan->rs_cbuf);
-
 			LockBuffer(scan->rs_cbuf, BUFFER_LOCK_UNLOCK);
 
 			if (!indexIt)
@@ -1615,41 +1608,31 @@ reindex_index(Oid indexId)
 {
 	Relation	iRel,
 				heapRelation;
-	IndexInfo  *indexInfo;
 	Oid			heapId;
 	bool		inplace;
 
 	/*
-	 * Open our index relation and get an exclusive lock on it.
-	 *
-	 * Note: for REINDEX INDEX, doing this before opening the parent heap
-	 * relation means there's a possibility for deadlock failure against
-	 * another xact that is doing normal accesses to the heap and index.
-	 * However, it's not real clear why you'd be wanting to do REINDEX INDEX
-	 * on a table that's in active use, so I'd rather have the protection of
-	 * making sure the index is locked down.  In the REINDEX TABLE and
-	 * REINDEX DATABASE cases, there is no problem because caller already
-	 * holds exclusive lock on the parent table.
+	 * Open and lock the parent heap relation.  ShareLock is sufficient
+	 * since we only need to be sure no schema or data changes are going on.
+	 */
+	heapId = IndexGetRelation(indexId);
+	heapRelation = heap_open(heapId, ShareLock);
+
+	/*
+	 * Open the target index relation and get an exclusive lock on it,
+	 * to ensure that no one else is touching this particular index.
 	 */
 	iRel = index_open(indexId);
 	LockRelation(iRel, AccessExclusiveLock);
 
-	/* Get OID of index's parent table */
-	heapId = iRel->rd_index->indrelid;
-
-	/* Open and lock the parent heap relation */
-	heapRelation = heap_open(heapId, AccessExclusiveLock);
-
-	SetReindexProcessing(heapId, indexId);
-
 	/*
 	 * If it's a shared index, we must do inplace processing (because we
 	 * have no way to update relfilenode in other databases).  Otherwise
 	 * we can do it the normal transaction-safe way.
 	 *
 	 * Since inplace processing isn't crash-safe, we only allow it in a
-	 * standalone backend.  (In the REINDEX TABLE and REINDEX DATABASE cases,
-	 * the caller should have detected this.)
+	 * standalone backend.	(In the REINDEX TABLE and REINDEX DATABASE
+	 * cases, the caller should have detected this.)
 	 */
 	inplace = iRel->rd_rel->relisshared;
 
@@ -1659,63 +1642,77 @@ reindex_index(Oid indexId)
 				 errmsg("shared index \"%s\" can only be reindexed in stand-alone mode",
 						RelationGetRelationName(iRel))));
 
-	/* Fetch info needed for index_build */
-	indexInfo = BuildIndexInfo(iRel);
-
-	if (inplace)
+	PG_TRY();
 	{
-		/*
-		 * Release any buffers associated with this index.	If they're
-		 * dirty, they're just dropped without bothering to flush to disk.
-		 */
-		DropRelationBuffers(iRel);
+		IndexInfo  *indexInfo;
 
-		/* Now truncate the actual data and set blocks to zero */
-		smgrtruncate(DEFAULT_SMGR, iRel, 0);
-		iRel->rd_nblocks = 0;
-		iRel->rd_targblock = InvalidBlockNumber;
+		/* Suppress use of the target index while rebuilding it */
+		SetReindexProcessing(heapId, indexId);
+
+		/* Fetch info needed for index_build */
+		indexInfo = BuildIndexInfo(iRel);
+
+		if (inplace)
+		{
+			/*
+			 * Release any buffers associated with this index.	If they're
+			 * dirty, they're just dropped without bothering to flush to
+			 * disk.
+			 */
+			DropRelationBuffers(iRel);
+
+			/* Now truncate the actual data */
+			RelationTruncate(iRel, 0);
+		}
+		else
+		{
+			/*
+			 * We'll build a new physical relation for the index.
+			 */
+			setNewRelfilenode(iRel);
+		}
+
+		/* Initialize the index and rebuild */
+		index_build(heapRelation, iRel, indexInfo);
+
+		/*
+		 * index_build will close both the heap and index relations (but
+		 * not give up the locks we hold on them).	So we're done.
+		 */
 	}
-	else
+	PG_CATCH();
 	{
-		/*
-		 * We'll build a new physical relation for the index.
-		 */
-		setNewRelfilenode(iRel);
+		/* Make sure flag gets cleared on error exit */
+		ResetReindexProcessing();
+		PG_RE_THROW();
 	}
-
-	/* Initialize the index and rebuild */
-	index_build(heapRelation, iRel, indexInfo);
-
-	/*
-	 * index_build will close both the heap and index relations (but not
-	 * give up the locks we hold on them).	So we're done.
-	 */
-	SetReindexProcessing(InvalidOid, InvalidOid);
+	PG_END_TRY();
+	ResetReindexProcessing();
 }
 
 /*
  * reindex_relation - This routine is used to recreate all indexes
- * of a relation (and its toast relation too, if any).
+ * of a relation (and optionally its toast relation too, if any).
  *
- * Returns true if any indexes were rebuilt.
+ * Returns true if any indexes were rebuilt.  Note that a
+ * CommandCounterIncrement will occur after each index rebuild.
  */
 bool
-reindex_relation(Oid relid)
+reindex_relation(Oid relid, bool toast_too)
 {
 	Relation	rel;
 	Oid			toast_relid;
 	bool		is_pg_class;
 	bool		result;
 	List	   *indexIds,
-			   *doneIndexes,
-			   *indexId;
+			   *doneIndexes;
+	ListCell   *indexId;
 
 	/*
-	 * Ensure to hold an exclusive lock throughout the transaction. The
-	 * lock could perhaps be less intensive (in the non-overwrite case)
-	 * but for now it's AccessExclusiveLock for simplicity.
+	 * Open and lock the relation.  ShareLock is sufficient since we only
+	 * need to prevent schema and data changes in it.
 	 */
-	rel = heap_open(relid, AccessExclusiveLock);
+	rel = heap_open(relid, ShareLock);
 
 	toast_relid = rel->rd_rel->reltoastrelid;
 
@@ -1728,13 +1725,13 @@ reindex_relation(Oid relid)
 
 	/*
 	 * reindex_index will attempt to update the pg_class rows for the
-	 * relation and index.  If we are processing pg_class itself, we
-	 * want to make sure that the updates do not try to insert index
-	 * entries into indexes we have not processed yet.  (When we are
-	 * trying to recover from corrupted indexes, that could easily
-	 * cause a crash.)  We can accomplish this because CatalogUpdateIndexes
-	 * will use the relcache's index list to know which indexes to update.
-	 * We just force the index list to be only the stuff we've processed.
+	 * relation and index.	If we are processing pg_class itself, we want
+	 * to make sure that the updates do not try to insert index entries
+	 * into indexes we have not processed yet.	(When we are trying to
+	 * recover from corrupted indexes, that could easily cause a crash.)
+	 * We can accomplish this because CatalogUpdateIndexes will use the
+	 * relcache's index list to know which indexes to update. We just
+	 * force the index list to be only the stuff we've processed.
 	 *
 	 * It is okay to not insert entries into the indexes we have not
 	 * processed yet because all of this is transaction-safe.  If we fail
@@ -1749,7 +1746,7 @@ reindex_relation(Oid relid)
 	/* Reindex all the indexes. */
 	foreach(indexId, indexIds)
 	{
-		Oid		indexOid = lfirsto(indexId);
+		Oid			indexOid = lfirst_oid(indexId);
 
 		if (is_pg_class)
 			RelationSetIndexList(rel, doneIndexes);
@@ -1759,7 +1756,7 @@ reindex_relation(Oid relid)
 		CommandCounterIncrement();
 
 		if (is_pg_class)
-			doneIndexes = lappendo(doneIndexes, indexOid);
+			doneIndexes = lappend_oid(doneIndexes, indexOid);
 	}
 
 	if (is_pg_class)
@@ -1773,11 +1770,11 @@ reindex_relation(Oid relid)
 	result = (indexIds != NIL);
 
 	/*
-	 * If the relation has a secondary toast rel, reindex that too while we
-	 * still hold the lock on the master table.
+	 * If the relation has a secondary toast rel, reindex that too while
+	 * we still hold the lock on the master table.
 	 */
-	if (toast_relid != InvalidOid)
-		result |= reindex_relation(toast_relid);
+	if (toast_too && OidIsValid(toast_relid))
+		result |= reindex_relation(toast_relid, false);
 
 	return result;
 }
