@@ -3,11 +3,11 @@
  * fd.c
  *	  Virtual file descriptor code.
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/storage/file/fd.c,v 1.102.2.3 2005/08/07 18:48:00 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/storage/file/fd.c,v 1.115.4.1 2005/08/07 18:47:38 tgl Exp $
  *
  * NOTES:
  *
@@ -47,21 +47,17 @@
 #include <fcntl.h>
 
 #include "miscadmin.h"
+#include "access/xact.h"
 #include "storage/fd.h"
 #include "storage/ipc.h"
 
 
-/* Filename components for OpenTemporaryFile */
-#define PG_TEMP_FILES_DIR "pgsql_tmp"
-#define PG_TEMP_FILE_PREFIX "pgsql_tmp"
-
-
 /*
  * We must leave some file descriptors free for system(), the dynamic loader,
  * and other code that tries to open files without consulting fd.c.  This
  * is the number left free.  (While we can be pretty sure we won't get
  * EMFILE, there's never any guarantee that we won't get ENFILE due to
- * other processes chewing up FDs.  So it's a bad idea to try to open files
+ * other processes chewing up FDs.	So it's a bad idea to try to open files
  * without consulting fd.c.  Nonetheless we cannot control all code.)
  *
  * Because this is just a fixed setting, we are effectively assuming that
@@ -99,7 +95,7 @@ int			max_files_per_process = 1000;
  * Note: the value of max_files_per_process is taken into account while
  * setting this variable, and so need not be tested separately.
  */
-static int	max_safe_fds = 32;			/* default if not changed */
+static int	max_safe_fds = 32;	/* default if not changed */
 
 
 /* Debugging.... */
@@ -127,6 +123,7 @@ typedef struct vfd
 {
 	signed short fd;			/* current FD, or VFD_CLOSED if none */
 	unsigned short fdstate;		/* bitflags for VFD's state */
+	SubTransactionId create_subid;	/* for TEMPORARY fds, creating subxact */
 	File		nextFree;		/* link to next free VFD, if in freelist */
 	File		lruMoreRecently;	/* doubly linked recency-of-use list */
 	File		lruLessRecently;
@@ -151,27 +148,34 @@ static Size SizeVfdCache = 0;
 static int	nfile = 0;
 
 /*
- * List of stdio FILEs opened with AllocateFile.
+ * List of stdio FILEs and <dirent.h> DIRs opened with AllocateFile
+ * and AllocateDir.
  *
- * Since we don't want to encourage heavy use of AllocateFile, it seems
- * OK to put a pretty small maximum limit on the number of simultaneously
- * allocated files.
+ * Since we don't want to encourage heavy use of AllocateFile or AllocateDir,
+ * it seems OK to put a pretty small maximum limit on the number of
+ * simultaneously allocated descs.
  */
-#define MAX_ALLOCATED_FILES  32
+#define MAX_ALLOCATED_DESCS  32
 
-static int	numAllocatedFiles = 0;
-static FILE *allocatedFiles[MAX_ALLOCATED_FILES];
+typedef enum
+{
+	AllocateDescFile,
+	AllocateDescDir
+} AllocateDescKind;
 
-/*
- * List of <dirent.h> DIRs opened with AllocateDir.
- *
- * Since we don't have heavy use of AllocateDir, it seems OK to put a pretty
- * small maximum limit on the number of simultaneously allocated dirs.
- */
-#define MAX_ALLOCATED_DIRS  10
+typedef struct
+{
+	AllocateDescKind kind;
+	union
+	{
+		FILE	   *file;
+		DIR		   *dir;
+	}			desc;
+	SubTransactionId create_subid;
+} AllocateDesc;
 
-static int	numAllocatedDirs = 0;
-static DIR *allocatedDirs[MAX_ALLOCATED_DIRS];
+static int	numAllocatedDescs = 0;
+static AllocateDesc allocatedDescs[MAX_ALLOCATED_DESCS];
 
 /*
  * Number of temporary files opened during the current session;
@@ -222,8 +226,9 @@ static void FreeVfd(File file);
 static int	FileAccess(File file);
 static File fileNameOpenFile(FileName fileName, int fileFlags, int fileMode);
 static char *filepath(const char *filename);
-static void AtProcExit_Files(void);
+static void AtProcExit_Files(int code, Datum arg);
 static void CleanupTempFiles(bool isProcExit);
+static void RemovePgTempFilesInDir(const char *tmpdirname);
 
 
 /*
@@ -285,7 +290,7 @@ count_usable_fds(int max_to_probe, int *usable_fds, int *already_open)
 	/* dup until failure or probe limit reached */
 	for (;;)
 	{
-		int		thisfd;
+		int			thisfd;
 
 		thisfd = dup(0);
 		if (thisfd < 0)
@@ -317,12 +322,12 @@ count_usable_fds(int max_to_probe, int *usable_fds, int *already_open)
 	pfree(fd);
 
 	/*
-	 * Return results.  usable_fds is just the number of successful dups.
-	 * We assume that the system limit is highestfd+1 (remember 0 is a legal
-	 * FD number) and so already_open is highestfd+1 - usable_fds.
+	 * Return results.	usable_fds is just the number of successful dups.
+	 * We assume that the system limit is highestfd+1 (remember 0 is a
+	 * legal FD number) and so already_open is highestfd+1 - usable_fds.
 	 */
 	*usable_fds = used;
-	*already_open = highestfd+1 - used;
+	*already_open = highestfd + 1 - used;
 }
 
 /*
@@ -336,11 +341,11 @@ set_max_safe_fds(void)
 	int			already_open;
 
 	/*
-	 * We want to set max_safe_fds to
-	 *			MIN(usable_fds, max_files_per_process - already_open)
-	 * less the slop factor for files that are opened without consulting
-	 * fd.c.  This ensures that we won't exceed either max_files_per_process
-	 * or the experimentally-determined EMFILE limit.
+	 * We want to set max_safe_fds to MIN(usable_fds,
+	 * max_files_per_process - already_open) less the slop factor for
+	 * files that are opened without consulting fd.c.  This ensures that
+	 * we won't exceed either max_files_per_process or the
+	 * experimentally-determined EMFILE limit.
 	 */
 	count_usable_fds(max_files_per_process,
 					 &usable_fds, &already_open);
@@ -471,7 +476,7 @@ LruDelete(File file)
 
 	/* close the file */
 	if (close(vfdP->fd))
-		elog(LOG, "failed to close \"%s\": %m",
+		elog(ERROR, "failed to close \"%s\": %m",
 			 vfdP->fileName);
 
 	--nfile;
@@ -499,6 +504,7 @@ Insert(File file)
 	DO_DB(_dump_lru());
 }
 
+/* returns 0 on success, -1 on re-open failure (with errno set) */
 static int
 LruInsert(File file)
 {
@@ -513,7 +519,7 @@ LruInsert(File file)
 
 	if (FileIsNotOpen(file))
 	{
-		while (nfile + numAllocatedFiles + numAllocatedDirs >= max_safe_fds)
+		while (nfile + numAllocatedDescs >= max_safe_fds)
 		{
 			if (!ReleaseLruFile())
 				break;
@@ -700,6 +706,7 @@ filepath(const char *filename)
 	return buf;
 }
 
+/* returns 0 on success, -1 on re-open failure (with errno set) */
 static int
 FileAccess(File file)
 {
@@ -772,7 +779,7 @@ fileNameOpenFile(FileName fileName,
 	file = AllocateVfd();
 	vfdP = &VfdCache[file];
 
-	while (nfile + numAllocatedFiles + numAllocatedDirs >= max_safe_fds)
+	while (nfile + numAllocatedDescs >= max_safe_fds)
 	{
 		if (!ReleaseLruFile())
 			break;
@@ -889,7 +896,10 @@ OpenTemporaryFile(bool interXact)
 
 	/* Mark it for deletion at EOXact */
 	if (!interXact)
+	{
 		VfdCache[file].fdstate |= FD_XACT_TEMPORARY;
+		VfdCache[file].create_subid = GetCurrentSubTransactionId();
+	}
 
 	return file;
 }
@@ -916,7 +926,7 @@ FileClose(File file)
 
 		/* close the file */
 		if (close(vfdP->fd))
-			elog(LOG, "failed to close \"%s\": %m",
+			elog(ERROR, "failed to close \"%s\": %m",
 				 vfdP->fileName);
 
 		--nfile;
@@ -969,7 +979,10 @@ FileRead(File file, char *buffer, int amount)
 			   file, VfdCache[file].fileName,
 			   VfdCache[file].seekPos, amount, buffer));
 
-	FileAccess(file);
+	returnCode = FileAccess(file);
+	if (returnCode < 0)
+		return returnCode;
+
 	returnCode = read(VfdCache[file].fd, buffer, amount);
 	if (returnCode > 0)
 		VfdCache[file].seekPos += returnCode;
@@ -990,7 +1003,9 @@ FileWrite(File file, char *buffer, int amount)
 			   file, VfdCache[file].fileName,
 			   VfdCache[file].seekPos, amount, buffer));
 
-	FileAccess(file);
+	returnCode = FileAccess(file);
+	if (returnCode < 0)
+		return returnCode;
 
 	errno = 0;
 	returnCode = write(VfdCache[file].fd, buffer, amount);
@@ -1007,9 +1022,28 @@ FileWrite(File file, char *buffer, int amount)
 	return returnCode;
 }
 
+int
+FileSync(File file)
+{
+	int			returnCode;
+
+	Assert(FileIsValid(file));
+
+	DO_DB(elog(LOG, "FileSync: %d (%s)",
+			   file, VfdCache[file].fileName));
+
+	returnCode = FileAccess(file);
+	if (returnCode < 0)
+		return returnCode;
+
+	return pg_fsync(VfdCache[file].fd);
+}
+
 long
 FileSeek(File file, long offset, int whence)
 {
+	int			returnCode;
+
 	Assert(FileIsValid(file));
 
 	DO_DB(elog(LOG, "FileSeek: %d (%s) %ld %ld %d",
@@ -1029,8 +1063,11 @@ FileSeek(File file, long offset, int whence)
 				VfdCache[file].seekPos += offset;
 				break;
 			case SEEK_END:
-				FileAccess(file);
-				VfdCache[file].seekPos = lseek(VfdCache[file].fd, offset, whence);
+				returnCode = FileAccess(file);
+				if (returnCode < 0)
+					return returnCode;
+				VfdCache[file].seekPos = lseek(VfdCache[file].fd,
+											   offset, whence);
 				break;
 			default:
 				elog(ERROR, "invalid whence: %d", whence);
@@ -1045,14 +1082,17 @@ FileSeek(File file, long offset, int whence)
 				if (offset < 0)
 					elog(ERROR, "invalid seek offset: %ld", offset);
 				if (VfdCache[file].seekPos != offset)
-					VfdCache[file].seekPos = lseek(VfdCache[file].fd, offset, whence);
+					VfdCache[file].seekPos = lseek(VfdCache[file].fd,
+												   offset, whence);
 				break;
 			case SEEK_CUR:
 				if (offset != 0 || VfdCache[file].seekPos == FileUnknownPos)
-					VfdCache[file].seekPos = lseek(VfdCache[file].fd, offset, whence);
+					VfdCache[file].seekPos = lseek(VfdCache[file].fd,
+												   offset, whence);
 				break;
 			case SEEK_END:
-				VfdCache[file].seekPos = lseek(VfdCache[file].fd, offset, whence);
+				VfdCache[file].seekPos = lseek(VfdCache[file].fd,
+											   offset, whence);
 				break;
 			default:
 				elog(ERROR, "invalid whence: %d", whence);
@@ -1086,7 +1126,10 @@ FileTruncate(File file, long offset)
 	DO_DB(elog(LOG, "FileTruncate %d (%s)",
 			   file, VfdCache[file].fileName));
 
-	FileAccess(file);
+	returnCode = FileAccess(file);
+	if (returnCode < 0)
+		return returnCode;
+
 	returnCode = ftruncate(VfdCache[file].fd, (size_t) offset);
 	return returnCode;
 }
@@ -1114,24 +1157,29 @@ AllocateFile(char *name, char *mode)
 {
 	FILE	   *file;
 
-	DO_DB(elog(LOG, "AllocateFile: Allocated %d", numAllocatedFiles));
+	DO_DB(elog(LOG, "AllocateFile: Allocated %d (%s)",
+			   numAllocatedDescs, name));
 
 	/*
-	 * The test against MAX_ALLOCATED_FILES prevents us from overflowing
-	 * allocatedFiles[]; the test against max_safe_fds prevents AllocateFile
-	 * from hogging every one of the available FDs, which'd lead to infinite
-	 * looping.
+	 * The test against MAX_ALLOCATED_DESCS prevents us from overflowing
+	 * allocatedFiles[]; the test against max_safe_fds prevents
+	 * AllocateFile from hogging every one of the available FDs, which'd
+	 * lead to infinite looping.
 	 */
-	if (numAllocatedFiles >= MAX_ALLOCATED_FILES ||
-		numAllocatedFiles + numAllocatedDirs >= max_safe_fds - 1)
+	if (numAllocatedDescs >= MAX_ALLOCATED_DESCS ||
+		numAllocatedDescs >= max_safe_fds - 1)
 		elog(ERROR, "too many private files demanded");
 
 TryAgain:
 	if ((file = fopen(name, mode)) != NULL)
 	{
-		allocatedFiles[numAllocatedFiles] = file;
-		numAllocatedFiles++;
-		return file;
+		AllocateDesc *desc = &allocatedDescs[numAllocatedDescs];
+
+		desc->kind = AllocateDescFile;
+		desc->desc.file = file;
+		desc->create_subid = GetCurrentSubTransactionId();
+		numAllocatedDescs++;
+		return desc->desc.file;
 	}
 
 	if (errno == EMFILE || errno == ENFILE)
@@ -1150,27 +1198,64 @@ TryAgain:
 	return NULL;
 }
 
-void
+/*
+ * Free an AllocateDesc of either type.
+ *
+ * The argument *must* point into the allocatedDescs[] array.
+ */
+static int
+FreeDesc(AllocateDesc *desc)
+{
+	int			result;
+
+	/* Close the underlying object */
+	switch (desc->kind)
+	{
+		case AllocateDescFile:
+			result = fclose(desc->desc.file);
+			break;
+		case AllocateDescDir:
+			result = closedir(desc->desc.dir);
+			break;
+		default:
+			elog(ERROR, "AllocateDesc kind not recognized");
+			result = 0;			/* keep compiler quiet */
+			break;
+	}
+
+	/* Compact storage in the allocatedDescs array */
+	numAllocatedDescs--;
+	*desc = allocatedDescs[numAllocatedDescs];
+
+	return result;
+}
+
+/*
+ * Close a file returned by AllocateFile.
+ *
+ * Note we do not check fclose's return value --- it is up to the caller
+ * to handle close errors.
+ */
+int
 FreeFile(FILE *file)
 {
 	int			i;
 
-	DO_DB(elog(LOG, "FreeFile: Allocated %d", numAllocatedFiles));
+	DO_DB(elog(LOG, "FreeFile: Allocated %d", numAllocatedDescs));
 
 	/* Remove file from list of allocated files, if it's present */
-	for (i = numAllocatedFiles; --i >= 0;)
+	for (i = numAllocatedDescs; --i >= 0;)
 	{
-		if (allocatedFiles[i] == file)
-		{
-			numAllocatedFiles--;
-			allocatedFiles[i] = allocatedFiles[numAllocatedFiles];
-			break;
-		}
-	}
-	if (i < 0)
-		elog(WARNING, "file passed to FreeFile was not obtained from AllocateFile");
+		AllocateDesc *desc = &allocatedDescs[i];
 
-	fclose(file);
+		if (desc->kind == AllocateDescFile && desc->desc.file == file)
+			return FreeDesc(desc);
+	}
+
+	/* Only get here if someone passes us a file not in allocatedDescs */
+	elog(WARNING, "file passed to FreeFile was not obtained from AllocateFile");
+
+	return fclose(file);
 }
 
 
@@ -1185,26 +1270,31 @@ FreeFile(FILE *file)
 DIR *
 AllocateDir(const char *dirname)
 {
-	DIR	   *dir;
+	DIR		   *dir;
 
-	DO_DB(elog(LOG, "AllocateDir: Allocated %d", numAllocatedDirs));
+	DO_DB(elog(LOG, "AllocateDir: Allocated %d (%s)",
+			   numAllocatedDescs, dirname));
 
 	/*
-	 * The test against MAX_ALLOCATED_DIRS prevents us from overflowing
-	 * allocatedDirs[]; the test against max_safe_fds prevents AllocateDir
-	 * from hogging every one of the available FDs, which'd lead to infinite
-	 * looping.
+	 * The test against MAX_ALLOCATED_DESCS prevents us from overflowing
+	 * allocatedDescs[]; the test against max_safe_fds prevents
+	 * AllocateDir from hogging every one of the available FDs, which'd
+	 * lead to infinite looping.
 	 */
-	if (numAllocatedDirs >= MAX_ALLOCATED_DIRS ||
-		numAllocatedDirs + numAllocatedFiles >= max_safe_fds - 1)
+	if (numAllocatedDescs >= MAX_ALLOCATED_DESCS ||
+		numAllocatedDescs >= max_safe_fds - 1)
 		elog(ERROR, "too many private dirs demanded");
 
 TryAgain:
 	if ((dir = opendir(dirname)) != NULL)
 	{
-		allocatedDirs[numAllocatedDirs] = dir;
-		numAllocatedDirs++;
-		return dir;
+		AllocateDesc *desc = &allocatedDescs[numAllocatedDescs];
+
+		desc->kind = AllocateDescDir;
+		desc->desc.dir = dir;
+		desc->create_subid = GetCurrentSubTransactionId();
+		numAllocatedDescs++;
+		return desc->desc.dir;
 	}
 
 	if (errno == EMFILE || errno == ENFILE)
@@ -1234,20 +1324,19 @@ FreeDir(DIR *dir)
 {
 	int			i;
 
-	DO_DB(elog(LOG, "FreeDir: Allocated %d", numAllocatedDirs));
+	DO_DB(elog(LOG, "FreeDir: Allocated %d", numAllocatedDescs));
 
 	/* Remove dir from list of allocated dirs, if it's present */
-	for (i = numAllocatedDirs; --i >= 0;)
+	for (i = numAllocatedDescs; --i >= 0;)
 	{
-		if (allocatedDirs[i] == dir)
-		{
-			numAllocatedDirs--;
-			allocatedDirs[i] = allocatedDirs[numAllocatedDirs];
-			break;
-		}
+		AllocateDesc *desc = &allocatedDescs[i];
+
+		if (desc->kind == AllocateDescDir && desc->desc.dir == dir)
+			return FreeDesc(desc);
 	}
-	if (i < 0)
-		elog(WARNING, "dir passed to FreeDir was not obtained from AllocateDir");
+
+	/* Only get here if someone passes us a dir not in allocatedDescs */
+	elog(WARNING, "dir passed to FreeDir was not obtained from AllocateDir");
 
 	return closedir(dir);
 }
@@ -1276,6 +1365,52 @@ closeAllVfds(void)
 	}
 }
 
+/*
+ * AtEOSubXact_Files
+ *
+ * Take care of subtransaction commit/abort.  At abort, we close temp files
+ * that the subtransaction may have opened.  At commit, we reassign the
+ * files that were opened to the parent subtransaction.
+ */
+void
+AtEOSubXact_Files(bool isCommit, SubTransactionId mySubid,
+				  SubTransactionId parentSubid)
+{
+	Index		i;
+
+	if (SizeVfdCache > 0)
+	{
+		Assert(FileIsNotOpen(0));		/* Make sure ring not corrupted */
+		for (i = 1; i < SizeVfdCache; i++)
+		{
+			unsigned short fdstate = VfdCache[i].fdstate;
+
+			if ((fdstate & FD_XACT_TEMPORARY) &&
+				VfdCache[i].create_subid == mySubid)
+			{
+				if (isCommit)
+					VfdCache[i].create_subid = parentSubid;
+				else if (VfdCache[i].fileName != NULL)
+					FileClose(i);
+			}
+		}
+	}
+
+	for (i = 0; i < numAllocatedDescs; i++)
+	{
+		if (allocatedDescs[i].create_subid == mySubid)
+		{
+			if (isCommit)
+				allocatedDescs[i].create_subid = parentSubid;
+			else
+			{
+				/* have to recheck the item after FreeDesc (ugly) */
+				FreeDesc(&allocatedDescs[i--]);
+			}
+		}
+	}
+}
+
 /*
  * AtEOXact_Files
  *
@@ -1297,7 +1432,7 @@ AtEOXact_Files(void)
  * Here, we want to clean up *all* temp files including interXact ones.
  */
 static void
-AtProcExit_Files(void)
+AtProcExit_Files(int code, Datum arg)
 {
 	CleanupTempFiles(true);
 }
@@ -1336,11 +1471,8 @@ CleanupTempFiles(bool isProcExit)
 		}
 	}
 
-	while (numAllocatedFiles > 0)
-		FreeFile(allocatedFiles[0]);
-
-	while (numAllocatedDirs > 0)
-		FreeDir(allocatedDirs[0]);
+	while (numAllocatedDescs > 0)
+		FreeDesc(&allocatedDescs[0]);
 }
 
 
@@ -1361,55 +1493,83 @@ RemovePgTempFiles(void)
 {
 	char		db_path[MAXPGPATH];
 	char		temp_path[MAXPGPATH];
-	char		rm_path[MAXPGPATH];
 	DIR		   *db_dir;
-	DIR		   *temp_dir;
 	struct dirent *db_de;
-	struct dirent *temp_de;
 
 	/*
-	 * Cycle through pg_tempfiles for all databases and remove old temp
-	 * files.
+	 * Cycle through pgsql_tmp directories for all databases and remove old
+	 * temp files.
 	 */
 	snprintf(db_path, sizeof(db_path), "%s/base", DataDir);
-	if ((db_dir = AllocateDir(db_path)) != NULL)
+	db_dir = AllocateDir(db_path);
+	if (db_dir == NULL)
 	{
-		while ((db_de = readdir(db_dir)) != NULL)
-		{
-			if (strcmp(db_de->d_name, ".") == 0 ||
-				strcmp(db_de->d_name, "..") == 0)
-				continue;
-
-			snprintf(temp_path, sizeof(temp_path),
-					 "%s/%s/%s",
-					 db_path, db_de->d_name,
-					 PG_TEMP_FILES_DIR);
-			if ((temp_dir = AllocateDir(temp_path)) != NULL)
-			{
-				while ((temp_de = readdir(temp_dir)) != NULL)
-				{
-					if (strcmp(temp_de->d_name, ".") == 0 ||
-						strcmp(temp_de->d_name, "..") == 0)
-						continue;
-
-					snprintf(rm_path, sizeof(temp_path),
-							 "%s/%s/%s/%s",
-							 db_path, db_de->d_name,
-							 PG_TEMP_FILES_DIR,
-							 temp_de->d_name);
-
-					if (strncmp(temp_de->d_name,
-								PG_TEMP_FILE_PREFIX,
-								strlen(PG_TEMP_FILE_PREFIX)) == 0)
-						unlink(rm_path);
-					else
-						elog(LOG,
-							 "unexpected file found in temporary-files directory: \"%s\"",
-							 rm_path);
-				}
-				FreeDir(temp_dir);
-			}
-		}
-		FreeDir(db_dir);
+		/* this really should not happen */
+		elog(LOG, "could not open directory \"%s\": %m", db_path);
+		return;
 	}
+
+	while ((db_de = readdir(db_dir)) != NULL)
+	{
+		if (strcmp(db_de->d_name, ".") == 0 ||
+			strcmp(db_de->d_name, "..") == 0)
+			continue;
+
+		snprintf(temp_path, sizeof(temp_path), "%s/%s/%s",
+				 db_path, db_de->d_name, PG_TEMP_FILES_DIR);
+		RemovePgTempFilesInDir(temp_path);
+	}
+
+	FreeDir(db_dir);
+
+	/*
+	 * In EXEC_BACKEND case there is a pgsql_tmp directory at the top
+	 * level of DataDir as well.
+	 */
+#ifdef EXEC_BACKEND
+	snprintf(temp_path, sizeof(temp_path), "%s/%s",
+			 DataDir, PG_TEMP_FILES_DIR);
+	RemovePgTempFilesInDir(temp_path);
+#endif
+}
+
+/* Process one pgsql_tmp directory for RemovePgTempFiles */
+static void
+RemovePgTempFilesInDir(const char *tmpdirname)
+{
+	DIR		   *temp_dir;
+	struct dirent *temp_de;
+	char		rm_path[MAXPGPATH];
+
+	temp_dir = AllocateDir(tmpdirname);
+	if (temp_dir == NULL)
+	{
+		/* anything except ENOENT is fishy */
+		if (errno != ENOENT)
+			elog(LOG,
+				 "could not open temporary-files directory \"%s\": %m",
+				 tmpdirname);
+		return;
+	}
+
+	while ((temp_de = readdir(temp_dir)) != NULL)
+	{
+		if (strcmp(temp_de->d_name, ".") == 0 ||
+			strcmp(temp_de->d_name, "..") == 0)
+			continue;
+
+		snprintf(rm_path, sizeof(rm_path), "%s/%s",
+				 tmpdirname, temp_de->d_name);
+
+		if (strncmp(temp_de->d_name,
+					PG_TEMP_FILE_PREFIX,
+					strlen(PG_TEMP_FILE_PREFIX)) == 0)
+			unlink(rm_path);	/* note we ignore any error */
+		else
+			elog(LOG,
+				 "unexpected file found in temporary-files directory: \"%s\"",
+				 rm_path);
+	}
+
+	FreeDir(temp_dir);
 }
