@@ -4,12 +4,12 @@
  *	  dynamic hash tables
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/utils/hash/dynahash.c,v 1.48 2003/08/19 01:13:41 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/hash/dynahash.c,v 1.58.4.1 2005/06/18 20:51:44 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -63,10 +63,10 @@
  */
 static void *DynaHashAlloc(Size size);
 static HASHSEGMENT seg_alloc(HTAB *hashp);
-static bool element_alloc(HTAB *hashp);
+static bool element_alloc(HTAB *hashp, int nelem);
 static bool dir_realloc(HTAB *hashp);
 static bool expand_table(HTAB *hashp);
-static bool hdefault(HTAB *hashp);
+static void hdefault(HTAB *hashp);
 static bool init_htab(HTAB *hashp, long nelem);
 static void hash_corrupted(HTAB *hashp);
 
@@ -85,6 +85,7 @@ DynaHashAlloc(Size size)
 }
 
 #define MEM_ALLOC		DynaHashAlloc
+#undef	MEM_FREE				/* already in windows header files */
 #define MEM_FREE		pfree
 
 
@@ -119,8 +120,6 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 
 	/* Initialize the hash header */
 	hashp = (HTAB *) MEM_ALLOC(sizeof(HTAB));
-	if (!hashp)
-		return NULL;
 	MemSet(hashp, 0, sizeof(HTAB));
 
 	hashp->tabname = (char *) MEM_ALLOC(strlen(tabname) + 1);
@@ -134,7 +133,7 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 	/*
 	 * If you don't specify a match function, it defaults to strncmp() if
 	 * you used string_hash (either explicitly or by default) and to
-	 * memcmp() otherwise.  (Prior to PostgreSQL 7.4, memcmp() was always
+	 * memcmp() otherwise.	(Prior to PostgreSQL 7.4, memcmp() was always
 	 * used.)
 	 */
 	if (flags & HASH_COMPARE)
@@ -144,11 +143,19 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 	else
 		hashp->match = memcmp;
 
+	/*
+	 * Similarly, the key-copying function defaults to strncpy() or memcpy().
+	 */
+	if (hashp->hash == string_hash)
+		hashp->keycopy = (HashCopyFunc) strncpy;
+	else
+		hashp->keycopy = memcpy;
+
 	if (flags & HASH_SHARED_MEM)
 	{
 		/*
 		 * ctl structure is preallocated for shared memory tables. Note
-		 * that HASH_DIRSIZE had better be set as well.
+		 * that HASH_DIRSIZE and HASH_ALLOC had better be set as well.
 		 */
 		hashp->hctl = info->hctl;
 		hashp->dir = info->dir;
@@ -174,11 +181,12 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 	{
 		hashp->hctl = (HASHHDR *) hashp->alloc(sizeof(HASHHDR));
 		if (!hashp->hctl)
-			return NULL;
+			ereport(ERROR,
+					(errcode(ERRCODE_OUT_OF_MEMORY),
+					 errmsg("out of memory")));
 	}
 
-	if (!hdefault(hashp))
-		return NULL;
+	hdefault(hashp);
 
 	hctl = hashp->hctl;
 #ifdef HASH_STATISTICS
@@ -220,25 +228,42 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 	{
 		/* remaining hash table structures live in child of given context */
 		hashp->hcxt = AllocSetContextCreate(CurrentDynaHashCxt,
-											"DynaHashTable",
+											tabname,
 											ALLOCSET_DEFAULT_MINSIZE,
 											ALLOCSET_DEFAULT_INITSIZE,
 											ALLOCSET_DEFAULT_MAXSIZE);
 		CurrentDynaHashCxt = hashp->hcxt;
 	}
 
+	/* Build the hash directory structure */
 	if (!init_htab(hashp, nelem))
 	{
 		hash_destroy(hashp);
-		return NULL;
+		elog(ERROR, "failed to initialize hash table");
 	}
+
+	/*
+	 * For a shared hash table, preallocate the requested number of elements.
+	 * This reduces problems with run-time out-of-shared-memory conditions.
+	 */
+	if (flags & HASH_SHARED_MEM)
+	{
+		if (!element_alloc(hashp, (int) nelem))
+		{
+			hash_destroy(hashp);
+			ereport(ERROR,
+					(errcode(ERRCODE_OUT_OF_MEMORY),
+					 errmsg("out of memory")));
+		}
+	}
+
 	return hashp;
 }
 
 /*
  * Set default HASHHDR parameters.
  */
-static bool
+static void
 hdefault(HTAB *hashp)
 {
 	HASHHDR    *hctl = hashp->hctl;
@@ -252,8 +277,6 @@ hdefault(HTAB *hashp)
 	hctl->nentries = 0;
 	hctl->nsegs = 0;
 
-	/* I added these MS. */
-
 	/* rather pointless defaults for key & entry size */
 	hctl->keysize = sizeof(char *);
 	hctl->entrysize = 2 * sizeof(char *);
@@ -263,8 +286,6 @@ hdefault(HTAB *hashp)
 
 	/* garbage collection for HASH_REMOVE */
 	hctl->freeList = NULL;
-
-	return true;
 }
 
 
@@ -630,7 +651,7 @@ hash_search(HTAB *hashp,
 			if (currBucket == NULL)
 			{
 				/* no free elements.  allocate another chunk of buckets */
-				if (!element_alloc(hashp))
+				if (!element_alloc(hashp, HASHELEMENT_ALLOC_INCR))
 					return NULL;	/* out of memory */
 				currBucket = hctl->freeList;
 				Assert(currBucket != NULL);
@@ -644,12 +665,12 @@ hash_search(HTAB *hashp,
 
 			/* copy key into record */
 			currBucket->hashvalue = hashvalue;
-			memcpy(ELEMENTKEY(currBucket), keyPtr, hctl->keysize);
+			hashp->keycopy(ELEMENTKEY(currBucket), keyPtr, hctl->keysize);
 
 			/* caller is expected to fill the data field on return */
 
 			/* Check if it is time to split the segment */
-			if (++hctl->nentries / (long) (hctl->max_bucket + 1) > hctl->ffactor)
+			if (++hctl->nentries / (long) (hctl->max_bucket + 1) >= hctl->ffactor)
 			{
 				/*
 				 * NOTE: failure to expand table is not a fatal error, it
@@ -805,9 +826,9 @@ expand_table(HTAB *hashp)
 
 	/*
 	 * Relocate records to the new bucket.	NOTE: because of the way the
-	 * hash masking is done in calc_bucket, only one old bucket can need to
-	 * be split at this point.	With a different way of reducing the hash
-	 * value, that might not be true!
+	 * hash masking is done in calc_bucket, only one old bucket can need
+	 * to be split at this point.  With a different way of reducing the
+	 * hash value, that might not be true!
 	 */
 	old_segnum = old_bucket >> hctl->sshift;
 	old_segndx = MOD(old_bucket, hctl->ssize);
@@ -897,7 +918,7 @@ seg_alloc(HTAB *hashp)
  * allocate some new elements and link them into the free list
  */
 static bool
-element_alloc(HTAB *hashp)
+element_alloc(HTAB *hashp, int nelem)
 {
 	HASHHDR    *hctl = hashp->hctl;
 	Size		elementSize;
@@ -909,13 +930,13 @@ element_alloc(HTAB *hashp)
 
 	CurrentDynaHashCxt = hashp->hcxt;
 	tmpElement = (HASHELEMENT *)
-		hashp->alloc(HASHELEMENT_ALLOC_INCR * elementSize);
+		hashp->alloc(nelem * elementSize);
 
 	if (!tmpElement)
 		return false;
 
 	/* link all the new entries into the freelist */
-	for (i = 0; i < HASHELEMENT_ALLOC_INCR; i++)
+	for (i = 0; i < nelem; i++)
 	{
 		tmpElement->link = hctl->freeList;
 		hctl->freeList = tmpElement;
