@@ -17,7 +17,7 @@
  *
  *
  * IDENTIFICATION
- *		$Header: /cvsroot/pgsql-server/src/bin/pg_dump/pg_backup_archiver.h,v 1.52 2003/10/03 20:10:59 tgl Exp $
+ *		$PostgreSQL: pgsql/src/bin/pg_dump/pg_backup_archiver.h,v 1.62.4.1 2005/01/25 22:44:47 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -27,13 +27,19 @@
 
 #include "postgres_fe.h"
 
-#include <stdio.h>
 #include <time.h>
-#include <errno.h>
 
+#include "pg_backup.h"
+
+#include "libpq-fe.h"
 #include "pqexpbuffer.h"
+
 #define LOBBUFSIZE 32768
 
+/*
+ * Note: zlib.h must be included *after* libpq-fe.h, because the latter may
+ * include ssl.h, which has a naming conflict with zlib.h.
+ */
 #ifdef HAVE_LIBZ
 #include <zlib.h>
 #define GZCLOSE(fh) gzclose(fh)
@@ -55,11 +61,8 @@ typedef struct _z_stream
 typedef z_stream *z_streamp;
 #endif
 
-#include "pg_backup.h"
-#include "libpq-fe.h"
-
 #define K_VERS_MAJOR 1
-#define K_VERS_MINOR 7
+#define K_VERS_MINOR 10
 #define K_VERS_REV 0
 
 /* Data block types */
@@ -76,7 +79,14 @@ typedef z_stream *z_streamp;
 #define K_VERS_1_6 (( (1 * 256 + 6) * 256 + 0) * 256 + 0)		/* Schema field in TOCs */
 #define K_VERS_1_7 (( (1 * 256 + 7) * 256 + 0) * 256 + 0)		/* File Offset size in
 																 * header */
-#define K_VERS_MAX (( (1 * 256 + 7) * 256 + 255) * 256 + 0)
+#define K_VERS_1_8 (( (1 * 256 + 8) * 256 + 0) * 256 + 0)		/* change interpretation
+																 * of ID numbers and
+																 * dependencies */
+#define K_VERS_1_9 (( (1 * 256 + 9) * 256 + 0) * 256 + 0)		/* add default_with_oids
+																 * tracking */
+#define K_VERS_1_10 (( (1 * 256 + 10) * 256 + 0) * 256 + 0)		/* add tablespace */
+
+#define K_VERS_MAX (( (1 * 256 + 10) * 256 + 255) * 256 + 0)
 
 /* No of BLOBs to restore in 1 TX */
 #define BLOB_BATCH_SIZE 100
@@ -114,8 +124,6 @@ typedef void (*PrintTocDataPtr) (struct _archiveHandle * AH, struct _tocEntry *
 
 typedef size_t (*CustomOutPtr) (struct _archiveHandle * AH, const void *buf, size_t len);
 
-typedef int (*TocSortCompareFn) (const void *te1, const void *te2);
-
 typedef enum _archiveMode
 {
 	archModeWrite,
@@ -133,7 +141,9 @@ typedef enum
 	SQL_SCAN = 0,
 	SQL_IN_SQL_COMMENT,
 	SQL_IN_EXT_COMMENT,
-	SQL_IN_QUOTE
+	SQL_IN_QUOTE,
+	SQL_IN_DOLLARTAG,
+	SQL_IN_DOLLARQUOTE
 } sqlparseState;
 
 typedef struct
@@ -143,8 +153,24 @@ typedef struct
 	char		lastChar;
 	char		quoteChar;
 	int			braceDepth;
+	PQExpBuffer tagBuf;
 } sqlparseInfo;
 
+typedef enum
+{
+	STAGE_NONE = 0,
+	STAGE_INITIALIZING,
+	STAGE_PROCESSING,
+	STAGE_FINALIZING
+} ArchiverStage;
+
+typedef enum
+{
+	REQ_SCHEMA = 1,
+	REQ_DATA = 2,
+	REQ_ALL = REQ_SCHEMA + REQ_DATA
+} teReqs;
+
 typedef struct _archiveHandle
 {
 	Archive		public;			/* Public part of archive */
@@ -153,6 +179,11 @@ typedef struct _archiveHandle
 	char		vrev;
 	int			version;		/* Conveniently formatted version */
 
+	char	   *archiveRemoteVersion;	/* When reading an archive,
+										 * the version of the dumped DB */
+	char	   *archiveDumpVersion;		/* When reading an archive,
+										 * the version of the dumper */
+
 	int			debugLevel;		/* Used for logging (currently only by
 								 * --verbose) */
 	size_t		intSize;		/* Size of an integer in the archive */
@@ -222,7 +253,6 @@ typedef struct _archiveHandle
 	int			createdBlobXref;	/* Flag */
 	int			blobCount;		/* # of blobs restored */
 
-	int			lastID;			/* Last internal ID for a TOC entry */
 	char	   *fSpec;			/* Archive File Spec */
 	FILE	   *FH;				/* General purpose file handle */
 	void	   *OF;
@@ -230,6 +260,8 @@ typedef struct _archiveHandle
 
 	struct _tocEntry *toc;		/* List of TOC entries */
 	int			tocCount;		/* Number of TOC entries */
+	DumpId		maxDumpId;		/* largest DumpId among all TOC entries */
+
 	struct _tocEntry *currToc;	/* Used when dumping data */
 	int			compression;	/* Compression requested on open */
 	ArchiveMode mode;			/* File mode - r or w */
@@ -241,45 +273,51 @@ typedef struct _archiveHandle
 	/* these vars track state to avoid sending redundant SET commands */
 	char	   *currUser;		/* current username */
 	char	   *currSchema;		/* current schema */
-	bool		chk_fn_bodies;	/* current state of check_function_bodies */
+	char	   *currTablespace;	/* current tablespace */
+	bool		currWithOids;	/* current default_with_oids setting */
 
 	void	   *lo_buf;
 	size_t		lo_buf_used;
 	size_t		lo_buf_size;
+
+	int			noTocComments;
+	ArchiverStage stage;
+	ArchiverStage lastErrorStage;
+	struct _tocEntry *currentTE;
+	struct _tocEntry *lastErrorTE;
 } ArchiveHandle;
 
 typedef struct _tocEntry
 {
 	struct _tocEntry *prev;
 	struct _tocEntry *next;
-	int			id;
-	int			hadDumper;		/* Archiver was passed a dumper routine
+	CatalogId	catalogId;
+	DumpId		dumpId;
+	bool		hadDumper;		/* Archiver was passed a dumper routine
 								 * (used in restore) */
 	char	   *tag;			/* index tag */
 	char	   *namespace;		/* null or empty string if not in a schema */
+	char	   *tablespace;		/* null if not in a tablespace; empty string
+								 * means use database default */
 	char	   *owner;
+	bool		withOids;		/* Used only by "TABLE" tags */
 	char	   *desc;
 	char	   *defn;
 	char	   *dropStmt;
 	char	   *copyStmt;
-	char	   *oid;			/* Oid of source of entry */
-	Oid			oidVal;			/* Value of above */
-	const char *((*depOid)[]);
-	Oid			maxDepOidVal;	/* Value of largest OID in deps */
-	Oid			maxOidVal;		/* Max of entry OID and max dep OID */
+	DumpId	   *dependencies;	/* dumpIds of objects this one depends on */
+	int			nDeps;			/* number of dependencies */
 
-	int			printed;		/* Indicates if entry defn has been dumped */
 	DataDumperPtr dataDumper;	/* Routine to dump data for object */
 	void	   *dataDumperArg;	/* Arg for above routine */
 	void	   *formatData;		/* TOC Entry data specific to file format */
-
-	int			_moved;			/* Marker used when rearranging TOC */
-
 } TocEntry;
 
 /* Used everywhere */
 extern const char *progname;
+
 extern void die_horribly(ArchiveHandle *AH, const char *modulename, const char *fmt,...) __attribute__((format(printf, 3, 4)));
+extern void warn_or_die_horribly(ArchiveHandle *AH, const char *modulename, const char *fmt,...) __attribute__((format(printf, 3, 4)));
 extern void write_msg(const char *modulename, const char *fmt,...) __attribute__((format(printf, 2, 3)));
 
 extern void WriteTOC(ArchiveHandle *AH);
@@ -290,7 +328,7 @@ extern void WriteToc(ArchiveHandle *AH);
 extern void ReadToc(ArchiveHandle *AH);
 extern void WriteDataChunks(ArchiveHandle *AH);
 
-extern int	TocIDRequired(ArchiveHandle *AH, int id, RestoreOptions *ropt);
+extern teReqs TocIDRequired(ArchiveHandle *AH, DumpId id, RestoreOptions *ropt);
 extern bool checkSeek(FILE *fp);
 
 /*
